import pytest

from core_service import task


def test_wrong_task_arguments():
    with pytest.raises(ValueError):
        task(sleep_interval=-1)
    with pytest.raises(ValueError):
        task(workers=0)



from typing import List


def requirements(*deps: List[str]):
    """Decorator marking service method as a source of nested services list.

    Marked method should return a list of service instances that should be instantiated.

    Optional prop_name may contain name of object property with dependent service.
    Such method will be invoked only after attributes will be available, they will be a service
    and this service will be running.
    """
    def wrapper(f):
        f.requirements_definition = True
        f.service_requirements = deps
        return f

    return wrapper


def task(periodic: bool = True, sleep_interval: float = .1, workers: int = 1):
    """Decorator defining Service method as service task.

    Task will be started and stopped with a service.

    Task is `periodic` by default. It means that after task execution finished
    it will be started again after sleeping for `sleep_interval` seconds.

    Multiple instances of the service task can be started in parallel.
    It is started in single instance by default but you can control this behavior
    using `workers` argument.
    """
    if workers < 1:
        raise ValueError("Number of service task workers should be gte 1")
    if sleep_interval < 0:
        raise ValueError("Sleeping interval should be gte 0")

    def wrapper(f):
        f.service_task = True
        f.service_task_definition = {
            'periodic': periodic,
            'sleep_interval': sleep_interval,
            'workers': workers,
        }
        return f

    return wrapper



"""Service container.

Any service can contain other services. Nested services will be started
and stopped together.
"""
import abc
import inspect
import logging
from typing import List

from .abstract import AbstractService
from .exceptions import ServiceStartupException

log = logging.getLogger(__name__)


class ServiceCollection:
    """Collection of services.

    Services in collection can be started, stopped and checked together
    with appropriate methods.

    Services are started in direct order and stopped in backward. Only
    already started services will be stopped if some of the service startup
    failed.
    """
    services: List[AbstractService]
    started_services: List[AbstractService]

    def __init__(self):
        self.services = []
        self.started_services = []

    def add(self, service: AbstractService):
        """Add service to collection.
        """
        self.services.append(service)

    async def healthcheck(self):
        """Check health of all services in collection.
        """
        for service in self.services:
            await service.healthcheck()

    async def start_all(self):
        """Start all services or rollback on failure.

        Services will be started in order they were added.
        """
        try:
            for service in self.services:
                try:
                    await service.start()
                    await service.healthcheck()
                except Exception as e:
                    log.exception("Exception while starting %s service", service)
                    raise ServiceStartupException from e
                self.started_services.append(service)
        except ServiceStartupException:
            log.error("Stopping services on startup failure")
            await self.stop_all()
            raise

    async def stop_all(self):
        """Stop all services in collection.

        Only started services will be stopped. Services will be stopped
        in reverse to startup order.
        """
        log.debug("Stopping nested services.")
        for service in reversed(self.started_services):
            try:
                await service.stop()
            except Exception:  # noqa
                log.exception("Fail to stop %s service.", service)
        else:
            log.debug("There are no services to stop.")
        log.debug("All nested services were stopped.")


class ServiceContainerMixin(AbstractService, abc.ABC):
    """Service container mixin.

    Add service collection and related functionality to BaseService.
    """
    _services: ServiceCollection

    def __init__(self):
        self._services = ServiceCollection()
        super().__init__()

    async def healthcheck(self):
        await super().healthcheck()
        await self._services.healthcheck()

    async def _start_nested_services(self):
        """Start nested services.

        Underlying collection will be filled.

        Startup order depend on the requirement method name defined on Service class
        and the defined dependencies.

        :raise RuntimeError: if startup order can't be resolved
        """
        loaded = set()
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        ordering_required = [name for name, method in members
                             if hasattr(method, "requirements_definition")]
        self.log.debug("Requirements will be gathered from %s",
                       ', '.join(ordering_required))
        while ordering_required:
            ordered_count = 0
            for name in ordering_required[:]:
                self.log.debug("Check %s", name)
                method = getattr(self, name)
                requirements = getattr(method, "service_requirements")
                if len(requirements) > 0 and not loaded.issuperset(requirements):
                    self.log.debug("Not enought requirements. Loaded: %s, Required: %s",
                                   loaded, requirements)
                    continue
                self.log.debug("Getting requirements from %s", name)
                services = await method()
                self.log.debug("Requirements from %s: %s", method, services)
                if not (services is None or isinstance(services, list)):
                    raise TypeError("Requirements method must return list or None. "
                                    "It returns %s (%s type) instead.",
                                    services, type(services))
                if services:
                    for service in services:
                        self._services.add(service)
                ordering_required.remove(name)
                ordered_count += 1
                loaded.add(name)
                self.log.debug("Nested service %s was loaded", name)
            if ordered_count == 0:
                raise RuntimeError(
                    "Can't resolve services dependencies "
                    "from %s" % ', '.join(ordering_required)
                )

        await self._services.start_all()

    async def _stop_nested_services(self):
        """Stop nested services in reverse order.
        """
        await self._services.stop_all()



import logging


class ServiceLoggerAdapter(logging.LoggerAdapter):
    """
    This example adapter expects the passed in dict-like object to have a
    'connid' key, whose value in brackets is prepended to the log message.
    """
    def process(self, msg, kwargs):
        return '[%s] %s' % (self.extra['service'], msg), kwargs



import pytest
import asyncio

from core_service import Service, task


class TestTaskService(Service):
    example_task_counter = 0

    @task(sleep_interval=0.1)
    async def example_task(self):
        self.example_task_counter += 1


@pytest.mark.asyncio
async def test_service_task_simple():
    service = TestTaskService()
    await service.start()
    await asyncio.sleep(0)
    assert len(service._tasks.tasks) == 1
    await service.stop()
    assert service.example_task_counter > 0


@pytest.mark.asyncio
async def test_service_task_with_exception():
    class FailTaskService(Service):
        @task()
        async def example_fail_task(self):
            raise Exception("Fail for example")

    service = FailTaskService()
    await service.start()
    await asyncio.sleep(0)

    assert service.running is False

    await service.stop()



"""Tasks container.
"""
import abc
import asyncio
import inspect
import logging
from typing import Awaitable, Callable, List

from .abstract import AbstractService
from .exceptions import UnexpectedTaskException, UnhealthyException

log = logging.getLogger(__name__)


class ServiceTask:
    """Service task wrapper.

    Wraps service task async method. Contains task execution parameters.
    """
    service: AbstractService
    #: task will be executed in infinity loop with sleep_interval between runs
    periodic: bool = True
    #: number of seconds between task executions
    sleep_interval: float = .1
    #: number of task instances running in parallel
    workers: int = 1

    def __init__(self,
                 service: AbstractService,
                 f: Callable[[], Awaitable],
                 periodic: bool = True,
                 sleep_interval: float = .1,
                 workers: int = 1):
        self.callable = f
        self.service = service
        self.periodic = periodic
        self.sleep_interval = sleep_interval
        self.workers = workers

    async def run(self):
        """Run task.
        """
        while not self.service.should_stop:
            await self.callable()
            await asyncio.sleep(self.sleep_interval)


class TasksCollection:
    tasks: List[asyncio.Task]

    def __init__(self):
        self.tasks = []

    def add(self, task: asyncio.Task):
        """Add task with definition to collection.
        """
        self.tasks.append(task)

    def check_all(self):
        """Check tasks health.

        Remove finished tasks from collection.

        Raise exception if some task failed and `allow_fail` is set to `False` for it.
        """
        for task in self.tasks:
            if task.done():
                log.debug("Task %s is done", task)
                try:
                    e = task.exception()
                    if e:
                        raise e
                except asyncio.CancelledError:  # pragma: nocover
                    log.debug("Task %s was cancelled", task)
                except Exception as e:  # noqa
                    raise UnexpectedTaskException() from e
                self.tasks.remove(task)
                log.debug("Remove finished task %s from collection", task)

    async def stop_all(self, raise_exceptions=False):
        """Stop all tasks in collection.

        Cancel all tasks. Will raise exceptions if `raise_exceptions` is `True`
        or log exception instead.
        """
        task_list = []
        for task in self.tasks:
            task.cancel()
            task_list.append(task)
        results = await asyncio.gather(*task_list, return_exceptions=not raise_exceptions)
        # log exceptions if not raised
        if not raise_exceptions:
            for i, r in enumerate(results):
                if isinstance(r, Exception):
                    task = self.tasks[i]
                    try:
                        raise r
                    except asyncio.CancelledError:  # pragma: nocover
                        log.debug("Task %s was cancelled", task)
                    except Exception:  # noqa
                        log.exception("Task %s stopped with exception",
                                      task.get_name())
        log.debug("Cancelled %i service tasks", len(self.tasks))


class TasksMixin(AbstractService, abc.ABC):
    """Tasks mixin for BaseService.
    """
    _tasks: TasksCollection

    def __init__(self):
        super().__init__()
        self._tasks = TasksCollection()

    async def healthcheck(self):
        await super().healthcheck()
        try:
            self._tasks.check_all()
        except UnexpectedTaskException as e:
            self.log.exception("Service tasks healthcheck failed with exception")
            raise UnhealthyException from e

    async def _start_service_tasks(self):
        """Start tasks defined on the service.
        """
        members = inspect.getmembers(self, predicate=inspect.isroutine)
        for name, method in members:
            if hasattr(method, "service_task"):
                self.log.debug("Service task %s found", method)
                service_task = ServiceTask(self, method, **method.service_task_definition)
                for i in range(service_task.workers):
                    task_name = ".".join([self.name, method.__name__, str(i)])
                    log.debug("Create task %s", task_name)
                    task = self.loop.create_task(service_task.run(), name=task_name)
                    self._tasks.add(task)

    async def _stop_service_tasks(self):
        """Cancel and await all managed service tasks.
        """
        await self._tasks.stop_all()



import abc
import asyncio
import logging
from typing import Optional

from .exceptions import UnhealthyException
from .logging import ServiceLoggerAdapter


class AbstractService(abc.ABC):
    running: bool = False
    should_stop: bool = False

    _loop: Optional[asyncio.AbstractEventLoop] = None
    _log: Optional[ServiceLoggerAdapter] = None

    @property
    def loop(self):
        """Event loop
        """
        if not self._loop:
            self._loop = asyncio.get_event_loop()
        return self._loop

    @property
    def name(self):
        """Service name.

        Default implementation return class name.

        Used in log output etc. Should include unique identifiers.
        """
        return self.__class__.__name__

    @property
    def log(self):
        """Service logger with service name under the `service` key of extra
        """
        if self._log is None:
            logger = logging.getLogger(self.__class__.__module__)
            self._log = ServiceLoggerAdapter(logger, extra={'service': self.name})
        return self._log

    @abc.abstractmethod
    async def start(self):
        """Asynchronous start method.
        """
        pass  # pragma: nocover

    @abc.abstractmethod
    async def stop(self):
        """Asynchronous stop method.

        Service should gracefully finish all its activity.
        """
        pass  # pragma: nocover

    async def healthcheck(self):
        """Healthcheck method.

        This method can be invoked to check if service is healthy and running.
        Should raise :py:class:`core_service.exceptions.UnhealthyException`
        if it is not healthy.

        Default implementation checks that `running` flag is `True`. You can overload
        this method to add additional checks.
        """
        if not self.running:
            raise UnhealthyException

    def __del__(self):
        if self.running:
            raise RuntimeError("Service %s is not stopped correctly" % self.name)



import asyncio
from typing import Optional

from .abstract import AbstractService
from .container import ServiceContainerMixin
from .exceptions import ServiceStartupException, UnhealthyException
from .tasks import TasksMixin


class Service(ServiceContainerMixin, TasksMixin, AbstractService):
    """Base service class.

    Your services should be inherited from this class.

    Service class provides a basic asynchronous service lifecycle methods.
    You should extend it for your needs.
    """
    _monitoring_task: Optional[asyncio.Task] = None
    #: interval in seconds to sleep between healthcheck runs
    _monitoring_interval: float = .1

    def __init__(self, *, loop=None, monitoring_interval: float = .1):
        self._loop = loop
        self._monitoring_interval = monitoring_interval
        super().__init__()

    async def start(self):
        """Start service.

        Set `running` flag to `True`, start service tasks, nested services
        and create monitoring task.

        You can override this method in your service implementation to apply custom
        start logic. But don't forget to invoke super implementation.
        """
        self.log.debug("Starting")
        self.running = True
        await self._start_service_tasks()
        try:
            await self._start_nested_services()
        except (ServiceStartupException, TypeError):
            self.log.exception("Failed to start nested service")
            self.running = False
            self.should_stop = True
            await self._stop_service_tasks()
            raise
        self._monitoring_task = self.loop.create_task(self.monitoring_task(),
                                                      name=f"{self.name}.monitoring_task")
        self.log.info("Service was started")

    async def stop(self):
        """Stop service.

        Set `should_stop` flag to `True`, `running` to `False` and start shutdown sequence.

        Nested services will be stopped first, service tasks will be cancelled than.

        You can override this method in your service implementation to apply custom
        start logic. But don't forget to invoke super implementation.
        """
        self.should_stop = True
        self.running = False
        if self._monitoring_task:
            self._monitoring_task.cancel()
        await self._stop_nested_services()
        await self._stop_service_tasks()

    async def monitoring_task(self):
        """Monitoring task.

        Started with a service. Run healthcheck periodically and force service
        to stop if it failed.
        """
        while not self.should_stop:
            try:
                await self.healthcheck()
            except UnhealthyException:
                self.log.exception("Healthcheck failed with exception")
                break
            except Exception:  # noqa
                self.log.exception("Service healthcheck failed with unexpected exception")
                break
            await asyncio.sleep(self._monitoring_interval)
        # terminate service on exit
        if not self.should_stop:
            await self.stop()



import pytest

from core_service import Service


class TestService(Service):
    pass


@pytest.mark.asyncio
async def test_simple(event_loop):
    service = TestService(loop=event_loop)
    await service.start()
    await service.stop()


@pytest.mark.asyncio
async def test_unantended_deletion():
    service = TestService()
    await service.start()
    with pytest.raises(RuntimeError):
        service.__del__()
    await service.stop()



from setuptools import setup, find_packages

setup(
    name='core-service',
    version='0.1.0',
    url='https://github.com/sarafanio/core-service.git',
    author='Sarafan Community',
    author_email='flu2020@pm.me',
    description='asyncio service microframework',
    packages=find_packages(),
    install_requires=[
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-asyncio',
            'pytest-cov',
            'pylama',
            'mypy',
            'Sphinx',
        ]
    }
)



import asyncio

import pytest

from core_service import Service, requirements, task
from core_service.exceptions import ServiceStartupException


class NestedService(Service):
    pass


class MainService(Service):
    def __init__(self, *, nested=None, loop=None):
        super().__init__(loop=loop)
        if nested is None:
            nested = NestedService
        self.nested = nested(loop=loop)

    @requirements()
    async def main_requirements(self):
        return [
            self.nested
        ]


@pytest.mark.asyncio
async def test_nested_service_lifecycle():

    service = MainService()

    await service.start()
    assert service.running
    assert service.nested.running

    await service.stop()
    assert not service.running
    assert not service.nested.running


@pytest.mark.asyncio
async def test_nested_service_startup_failure():
    class FailStartupService(Service):
        async def start(self):
            raise Exception("Example startup exception")

    service = MainService(nested=FailStartupService)
    with pytest.raises(ServiceStartupException):
        await service.start()


@pytest.mark.asyncio
async def test_nested_service_silent_startup_failure():
    class FailOnStartupService(Service):
        async def start(self):
            await super().start()
            self.running = False

    service = MainService(nested=FailOnStartupService)
    with pytest.raises(ServiceStartupException):
        await service.start()


@pytest.mark.asyncio
async def test_nested_service_task_failure():
    class FailService(Service):
        @task()
        async def fail_task(self):
            raise Exception

    service = MainService(nested=FailService)
    await service.start()
    await asyncio.sleep(0)
    assert not service.running
    assert not service.nested.running
    await service.stop()


@pytest.mark.asyncio
async def test_nested_service_stop_failure():
    class FailOnStopService(Service):
        async def stop(self):
            await super().stop()
            raise Exception("Example stop exception")

    service = MainService(nested=FailOnStopService)
    await service.start()
    await service.stop()
    # TODO: check log contains exception


@pytest.mark.asyncio
async def test_broken_healthcheck():
    """Healthcheck method of nested service is broken.

    ServiceStartupException should be raised on `start()`.
    """
    class BrokenHealthcheckService(Service):
        async def healthcheck(self):
            raise Exception("Broken healthcheck for testing")

    service = MainService(nested=BrokenHealthcheckService)
    with pytest.raises(ServiceStartupException):
        await service.start()


@pytest.mark.asyncio
async def test_nested_service_relation():
    class RequiredService(Service):
        pass

    class DependentService(Service):
        def __init__(self, *, required_service, loop=None):
            super().__init__(loop=loop)
            self.required_service = required_service

        async def start(self):
            if self.required_service.running is False:
                raise RuntimeError("RequiredService is not running")
            await super().start()

    class MainWiredService(Service):
        def __init__(self):
            super().__init__()
            self.required_service = RequiredService()

        @requirements('main_requirements')
        async def a_secondary_requirements(self):  # `a_` is used to became first
            return [
                DependentService(required_service=self.required_service)
            ]

        @requirements()
        async def main_requirements(self):
            return [
                self.required_service
            ]

        @requirements('main_requirements', 'a_secondary_requirements')
        async def third_party_requirements(self):
            return []

    service = MainWiredService()
    await service.start()
    await service.stop()


@pytest.mark.asyncio
async def test_wrong_requirements_output():
    class WrongRequirementsService(Service):
        @requirements()
        async def wrong_requirements_service(self):
            return 1

    service = WrongRequirementsService()
    with pytest.raises(TypeError):
        await service.start()


@pytest.mark.asyncio
async def test_unresolvable_requirements():
    class UnresolvableService(Service):
        @requirements('second_reqs')
        async def first_reqs(self):
            return []

        @requirements('first_reqs')
        async def second_reqs(self):
            return []

    service = UnresolvableService()
    with pytest.raises(RuntimeError):
        await service.start()



import asyncio

import pytest

from core_service.tasks import TasksCollection


async def example_task():
    return True


@pytest.mark.asyncio
async def test_finished_task_removed(event_loop):
    collection = TasksCollection()
    collection.add(event_loop.create_task(example_task()))
    await asyncio.sleep(0)
    collection.check_all()
    assert len(collection.tasks) == 0
